// File: Gconvert.cpp
//-----------------------------------------------------------------------------
// Gconvert v0.2 -- A simple tool to convert different graph formats.
// Copyright (C) 2004-2005 Pascal Pons.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//-----------------------------------------------------------------------------
// Author   : Pascal Pons 
// Email    : pons@liafa.jussieu.fr
// Web page : http://liafa.jussieu.fr/~pons/
// Location : Paris, France
// Time	    : December 2005
//-----------------------------------------------------------------------------
// see readme.txt for more details

#include "graph.h"
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <fstream>
//#include <cmath>

using namespace std;

void info() {
  cerr << "Gconvert v0.2 -- A simple tool to convert different graph formats." << endl;
  cerr << "Copyright (C) 2004-2005 Pascal Pons." << endl << endl;
}

void help(char* prog_name) {
  info();
  cerr << "usage: " << endl << prog_name << " input_format output_format [input_file] [-o output_file] [options]" << endl << endl;
  cerr << "input_format and output_format must be one of the following :" << endl;
  cerr << "\tAEL : Adjacency Edge List" << endl;
  cerr << "\tAAL : Adjacency Arc List" << endl;
  cerr << "\tEL  : Edge List" << endl;
  cerr << "\tDEL : Dynamic Edge List" << endl;
  cerr << "\tAL  : Arc List" << endl;
  cerr << "\tDAL : Dynamic Arc List" << endl;
  cerr << "\tWEL : Weighted Edge List" << endl;
  cerr << "\tDWEL: Dynamic Weighted Edge List" << endl;
  cerr << "\tWAL : Weighted Arc List" << endl;
  cerr << "\tDWAL: Dynamic Weighted Arc List" << endl; 
  cerr << "\tCL  : Clique List (input only)" << endl; 
  cerr << "\tWCL : Weighted Clique List (input only)" << endl << endl;
  cerr << "input_file: read the network from this file, stdin if not specified" << endl;
  cerr << "output_file: write the output network in this file, stdout if not specified" << endl << endl;
  cerr << "options:" << endl;
  cerr << "-mxxx\tthe multi edge policy :" << endl;
  cerr << "\t-madd   : the weight is the sum (default for static output)" << endl;
  cerr << "\t-mmulti : multi edges are allowed" << endl;
  cerr << "\t-maddexp: add and edges has an exponential decrease rate" << endl;
  cerr << "\t-madlin : add and edges has linear decrease rate" << endl;
  cerr << "-lxxx\tset the characteristic time to xxx (for dynamic graphs)" << endl;
  cerr << "-txxx\tset the output time to xxx (for convertion from dynamic to static)" << endl;
  cerr << "-dxxx\tspecify the characters that are delimiters in xxx (default is \" \\t\")" << endl;
  cerr << "-n\tnumber the vertices from 0 to nb_vertices-1" << endl;
  cerr << "-i [file] make an index of the numbers given by -n option" << endl;
  cerr << "-c\tkeep the largest connected component" << endl;
  cerr << "-q\t(quiet) do not display any message" << endl;
  cerr << "-h\t(help) print this help" << endl;
  exit(0);
}

int main(int argc, char** argv)
{
  bool quiet = false;
  char* output_file = 0;
  char* input_file = 0;
  char* delimiters = " \t";
  Format input_format = ADJACENCY_EDGE_LIST;
  Format output_format = ADJACENCY_EDGE_LIST;
  Multi_edge_policy multi_edge_policy = ADD;
  bool output_real_name = true;
  bool index = false;
  bool components = false;
  char* index_file = 0;
  float output_time = 0.;
  float caracteristic_time = 0.;

  if(argc < 2) help(argv[0]);

  string str;
  
  str = argv[1];
  for(unsigned int i = 0; i < str.length(); i++) str[i] = toupper(str[i]);
  if(str == "AEL") input_format = ADJACENCY_EDGE_LIST;
  else if(str == "AAL") input_format = ADJACENCY_ARC_LIST;
  else if(str == "EL") input_format = EDGE_LIST;
  else if(str == "DEL") input_format = DYNAMIC_EDGE_LIST;
  else if(str == "AL") input_format = ARC_LIST;
  else if(str == "DAL") input_format = DYNAMIC_ARC_LIST;
  else if(str == "WEL") input_format = WEIGHTED_EDGE_LIST;
  else if(str == "DWEL") input_format = DYNAMIC_WEIGHTED_EDGE_LIST;
  else if(str == "WAL") input_format = WEIGHTED_ARC_LIST;
  else if(str == "DWAL") input_format = DYNAMIC_WEIGHTED_ARC_LIST;
  else if(str == "CL") input_format = CLIQUE;
  else if(str == "WCL") input_format = WEIGHTED_CLIQUE;
  else help(argv[0]);

  str = argv[2];
  for(unsigned int i = 0; i < str.length(); i++) str[i] = toupper(str[i]);
  if(str == "AEL") output_format = ADJACENCY_EDGE_LIST;
  else if(str == "AAL") output_format = ADJACENCY_ARC_LIST;
  else if(str == "EL") output_format = EDGE_LIST;
  else if(str == "DEL") output_format = DYNAMIC_EDGE_LIST;
  else if(str == "AL") output_format = ARC_LIST;
  else if(str == "DAL") output_format = DYNAMIC_ARC_LIST;
  else if(str == "WEL") output_format = WEIGHTED_EDGE_LIST;
  else if(str == "DWEL") output_format = DYNAMIC_WEIGHTED_EDGE_LIST;
  else if(str == "WAL") output_format = WEIGHTED_ARC_LIST;
  else if(str == "DWAL") output_format = DYNAMIC_WEIGHTED_ARC_LIST;
  else help(argv[0]);
    
  for (int i = 3; i < argc; i++)  
    if(argv[i][0] == '-') {
      switch(argv[i][1]) {
	case 'n':
	  output_real_name = false;
	  break;
	  
	case 'i':
	  index = true;
	  if(argv[i][2] != 0) help(argv[0]);
	  if(i+1 < argc && argv[i+1][0] != '-') {
	    if(!index_file) index_file = argv[++i];
	    else help(argv[0]);
	  }
	  break;
	  
	case 'd':
	  delimiters = argv[i]+2;
	  break;
	  
	case 'q':
	  if(argv[i][2] != 0) help(argv[0]);
	  quiet = true;
	  break;
	  
	case 'o':
	  if(argv[i][2] != 0) help(argv[0]);
	  if(++i < argc && !output_file)
	    output_file = argv[i];
	  else help(argv[0]);
	  if(output_file[0] == '-') help(argv[0]);
	  break;

	case 'm':
	  str = argv[i];
	  if(str == "-madd") multi_edge_policy = ADD;
	  else if(str == "-mmulti") multi_edge_policy = MULTI;
	  else if(str == "-maddexp") multi_edge_policy = ADD_EXP;
	  else if(str == "-maddlin") multi_edge_policy = ADD_LIN;
	  else help(argv[0]);
	  break;

	case 't':
	  output_time = atof(argv[i]+2);
	  if(output_time <= 0.) help(argv[0]);
	  break;
	  
	case 'l':
	  caracteristic_time = atof(argv[i]+2);
	  if(caracteristic_time <= 0.) help(argv[0]);
	  break;
	  
	case 'c':
	  if(argv[i][2] != 0) help(argv[0]);
	  components = true;
	  break;
	  
	default:
	  help(argv[0]);
      }
    }
    else {
      if(!input_file) input_file = argv[i];
      else help(argv[0]);
    }

  if(!quiet) info();


  if((output_format == DYNAMIC_EDGE_LIST || output_format == DYNAMIC_ARC_LIST || output_format == DYNAMIC_WEIGHTED_EDGE_LIST || output_format == DYNAMIC_WEIGHTED_ARC_LIST) && (input_format != DYNAMIC_EDGE_LIST && input_format != DYNAMIC_ARC_LIST && input_format != DYNAMIC_WEIGHTED_EDGE_LIST && input_format != DYNAMIC_WEIGHTED_ARC_LIST)) {
    cerr << "Error: Unable to convert static graph into dynamic graph" << endl;
    exit(0);
  }

  if((output_format == DYNAMIC_EDGE_LIST || output_format == DYNAMIC_ARC_LIST || output_format == DYNAMIC_WEIGHTED_EDGE_LIST || output_format == DYNAMIC_WEIGHTED_ARC_LIST) && multi_edge_policy == ADD)
    multi_edge_policy = MULTI;


// read the graph from the correct input  
  Graph G = Graph(input_format, output_format, output_real_name, multi_edge_policy, output_time, caracteristic_time, delimiters, quiet);
  if(input_file) {
    ifstream fin;
    fin.open(input_file, ios::in);
    if(!fin.is_open()) {
      cerr << "unable to open file " << input_file << endl; 
      exit(0);
    }
    fin >> G;
    fin.close();
  }
  else cin >> G;

// compute the conected components if required


  if(components) G.compute_components();


//

  
  if(index) {
    if(index_file) {
      ofstream findex;
      findex.open(index_file, ios::out);
      if(!findex) {
	cerr << "unable to open the index file " << index_file << endl;
	exit(0);
      }
      findex << G.names;
      findex.close();
    }
    else cout << G.names << endl;
  }

  if(output_file) {
    ofstream fout;
    fout.open(output_file, ios::out);
    if(!fout) {
      cerr << "unable to open the output file " << output_file << endl;
      exit(0);
    }
    fout << G;
    fout.close();
  }
  else cout << G;

  return 0;
}
